/* (C) 2009-2010 Yana */

#define TESTMIPS32
#undef TESTMIPS64
#include "device.h"

void device_cons_put_hexadecimal(unsigned int hexadecimal) {
	unsigned int i;
	for (i = 0; i < 8; ++i) {
		device_cons_putchar("0123456789abcdef"[hexadecimal >> (28 - i * 4) & 0xf]);
	}
}

void device_cons_put_string(const char *string) {
	while (*string != 0) {
		device_cons_putchar(*string);
		++string;
	}
}

#include "yk0a_mips.h"

__inline__ static yk0a_uint32_t yk0a_interrupt_read_request(void) {
	return (yk0a_mips_read_cause() & YK0A_MIPS_CAUSE_IP_MASK) >> YK0A_MIPS_CAUSE_IP_SHIFT;
}

__inline__ static void yk0a_interrupt_write_mask(yk0a_uint32_t mask) {
	yk0a_mips_write_status((yk0a_mips_read_status() & ~YK0A_MIPS_STATUS_IM_MASK) | ((mask << YK0A_MIPS_STATUS_IM_SHIFT) & YK0A_MIPS_STATUS_IM_MASK));
}

__inline__ static void yk0a_interrupt_initialize(void) {
	yk0a_interrupt_write_mask(0);
}

#include <stddef.h>

#define GLUE_MIPS_ENTRYLO0 $2
#define GLUE_MIPS_ENTRYLO1 $3
#define GLUE_MIPS_BADVADDR $8
#define GLUE_MIPS_ENTRYHI $10

#define GLUE_MIPS_HAZARD_ENTRYLO0 ssnop
#define GLUE_MIPS_HAZARD_ENTRYLO1 ssnop
#define GLUE_MIPS_HAZARD_ENTRYHI ssnop

#define GLUE_MIPS_HAZARD_DI ssnop
#define GLUE_MIPS_HAZARD_EI ssnop
#define GLUE_MIPS_HAZARD_TLBWR ssnop

yk0a_uint32_t glue_mips_read_badvaddr(void) {
	yk0a_uint32_t badvaddr;
	__asm__ __volatile__("mfc0 %0, " YK0A_PREPROCESS_STRING(GLUE_MIPS_BADVADDR) : "=r"(badvaddr));
	return badvaddr;
}

void glue_mips_write_entrylo0(yk0a_uint32_t entrylo0) {
	__asm__ __volatile__("mtc0 %0, " YK0A_PREPROCESS_STRING(GLUE_MIPS_ENTRYLO0) "; " YK0A_PREPROCESS_STRING(GLUE_MIPS_HAZARD_ENTRYLO0) : : "r"(entrylo0));
}

void glue_mips_write_entrylo1(yk0a_uint32_t entrylo1) {
	__asm__ __volatile__("mtc0 %0, " YK0A_PREPROCESS_STRING(GLUE_MIPS_ENTRYLO1) "; " YK0A_PREPROCESS_STRING(GLUE_MIPS_HAZARD_ENTRYLO1) : : "r"(entrylo1));
}

void glue_mips_write_entryhi(yk0a_uint32_t entryhi) {
	__asm__ __volatile__("mtc0 %0, " YK0A_PREPROCESS_STRING(GLUE_MIPS_ENTRYHI) "; " YK0A_PREPROCESS_STRING(GLUE_MIPS_HAZARD_ENTRYHI) : : "r"(entryhi));
}

yk0a_uint32_t glue_mips_di(void) {
	yk0a_uint32_t status;
	__asm__ __volatile__("di %0; " YK0A_PREPROCESS_STRING(GLUE_MIPS_HAZARD_DI) : "=r"(status) : : "memory");
	return status;
}

yk0a_uint32_t glue_mips_ei(void) {
	yk0a_uint32_t status;
	__asm__ __volatile__("ei %0; " YK0A_PREPROCESS_STRING(GLUE_MIPS_HAZARD_EI) : "=r"(status) : : "memory");
	return status;
}

void glue_mips_tlbwr(void) {
	__asm__ __volatile__("tlbwr; " YK0A_PREPROCESS_STRING(GLUE_MIPS_HAZARD_TLBWR));
}

#include <signal.h>
#include <string.h>
#include "kernel.h"
#include "proc.h"

extern irq_hook_t *irq_handlers[NR_IRQ_VECTORS];

#define INTERRUPT_SIZE 8
#define KSEG_MASK 0x1ffff
#define REMOTE_PHYS 0xb0000
#define REMOTE_SIZE 0x10000 /* 256 MB */
#define REMOTE_VIR 0x10000
#define STACK_BOOT 12
#define STACK_FRAME 33

#if !defined(MINIX_PROCESS) || MINIX_PROCESS % CLICK_SIZE != 0
#error MINIX_PROCESS
#endif

void glue_update(void);

const yk0a_uint8_t glue_boot[] = "memory=80100000:00f00000\0rootdev=769";

volatile yk0a_uint32_t glue_interrupt_lock;
volatile yk0a_uint32_t glue_interrupt_mask;
volatile yk0a_uint32_t glue_interrupt_request;
volatile yk0a_uint32_t glue_nest;

yk0a_uint32_t glue_convert_page(struct proc *p, yk0a_uint32_t page) {
	if (page >= p->p_memmap[T].mem_vir && page < p->p_memmap[T].mem_vir + p->p_memmap[T].mem_len) {
		return ((page + (p->p_memmap[T].mem_phys & KSEG_MASK) - p->p_memmap[T].mem_vir) << 6) | 0x1a; /* C = 3, D = 0, V = 1, G = 0 */
	}
	if (page >= p->p_memmap[D].mem_vir && page < p->p_memmap[S].mem_vir + p->p_memmap[S].mem_len) {
		return ((page + (p->p_memmap[D].mem_phys & KSEG_MASK) - p->p_memmap[D].mem_vir) << 6) | 0x1e; /* C = 3, D = 1, V = 1, G = 0 */
	}
	if (iskernelp(p) || p->p_priv->s_farmem[0].in_use != 0) {
		if (page >= REMOTE_VIR && page < REMOTE_VIR + REMOTE_SIZE) {
			return ((page + (REMOTE_PHYS & KSEG_MASK) - REMOTE_VIR) << 6) | 0x16; /* C = 2, D = 1, V = 1, G = 0 */
		}
	}
	return 0;
}

void glue_exception(int signal) {
	if (iskernelp(proc_ptr)) {
		panic("task exception", signal);
	} else {
		cause_sig(proc_nr(proc_ptr), signal);
	}
}

yk0a_uint32_t glue_lock(void) {
	return glue_mips_di();
}

void glue_unlock(yk0a_uint32_t state) {
	if ((state & YK0A_MIPS_STATUS_IE) != 0) {
		glue_mips_ei();
	}
}

void glue_initialize(phys_bytes a) {
	++glue_nest;
	yk0a_interrupt_initialize();
	aout = a;
	cstart(0, 0, (phys_bytes)glue_boot >> 16, (phys_bytes)glue_boot, sizeof(glue_boot));
	main();
	glue_update();
	--glue_nest;
}

void glue_update(void) {
	if (next_ptr != NULL) {
		proc_ptr = next_ptr;
		next_ptr = NULL;
		glue_mips_write_entryhi(proc_ptr->p_nr + NR_TASKS);
	}
}

void glue_exception_interrupt_procedure(void) {
	yk0a_uint32_t badvaddr;
	yk0a_uint32_t cause;
	yk0a_uint32_t entrylo;
	yk0a_uint32_t i;
	if (glue_nest != 0) {
device_cons_put_string("{nest}\n");
while (1) {
}
	}
	++glue_nest;
	cause = yk0a_mips_read_cause();
	switch ((cause & YK0A_MIPS_CAUSE_EXCCODE_MASK) >> YK0A_MIPS_CAUSE_EXCCODE_SHIFT) {
	case 8: /* Sys */
		proc_ptr->p_reg.pc += 4; /* ignore YK0A_MIPS_CAUSE_BD */
#if defined(MINIX_ENABLE_YK0A)
		proc_ptr->p_reg.retreg = sys_call(proc_ptr->p_reg.gpr4, proc_ptr->p_reg.gpr5, (message *)proc_ptr->p_reg.gpr6, 0);
#else
		proc_ptr->p_reg.retreg = sys_call(proc_ptr->p_reg.retreg, proc_ptr->p_reg.gpr3, (message *)proc_ptr->p_reg.gpr4, 0);
#endif
		break;
	case 2: /* TLBL */
	case 3: /* TLBS */
		badvaddr = glue_mips_read_badvaddr();
		entrylo = glue_convert_page(proc_ptr, badvaddr / CLICK_SIZE);
		if (entrylo == 0) {
			glue_exception(SIGSEGV);
		} else {
			if (((badvaddr / CLICK_SIZE) & 1) == 0) {
				glue_mips_write_entrylo0(entrylo);
				glue_mips_write_entrylo1(glue_convert_page(proc_ptr, badvaddr / CLICK_SIZE + 1));
			} else {
				glue_mips_write_entrylo0(glue_convert_page(proc_ptr, badvaddr / CLICK_SIZE - 1));
				glue_mips_write_entrylo1(entrylo);
			}
			glue_mips_tlbwr();
		}
		break;
	case 1: /* Mod */
	case 4: /* AdEL */
	case 5: /* AdES */
		glue_exception(SIGSEGV);
		break;
	case 10: /* RI */
	case 11: /* CpU */
		glue_exception(SIGILL);
		break;
	default:
device_cons_put_string("{exception ");
device_cons_put_hexadecimal(cause);
device_cons_put_string("}\n");
while (1) {
}
		break;
	case 0: /* Int */
		glue_interrupt_request = yk0a_interrupt_read_request() & glue_interrupt_mask;
		for (i = INTERRUPT_SIZE - 1; i > 0; --i) {
			if ((1 << i & glue_interrupt_request) != 0) {
				glue_interrupt_mask &= ~(1 << i);
				yk0a_interrupt_write_mask(glue_interrupt_mask);
				glue_interrupt_request &= ~(1 << i);
				intr_handle(irq_handlers[i]);
				if (1 << i == DEVICE_INTERRUPT_RTC) {
					*(volatile yk0a_uint32_t *)DEVICE_RTC_INTERRUPT_ACK;
				}
				if (irq_actids[i] == 0) {
					glue_interrupt_mask |= 1 << i;
					yk0a_interrupt_write_mask(glue_interrupt_mask);
				}
			}
		}
		break;
	}
	glue_update();
	--glue_nest;
}

unsigned inb(U16_t _port) {
device_cons_put_string("{inb}\n");
while (1) {
}
}

unsigned inw(U16_t _port) {
device_cons_put_string("{inw}\n");
while (1) {
}
}

unsigned inl(U32_t _port) {
device_cons_put_string("{inl}\n");
while (1) {
}
}

void outb(U16_t _port, U8_t _value) {
device_cons_put_string("{outb}\n");
while (1) {
}
}

void outw(U16_t _port, U16_t _value) {
device_cons_put_string("{outw}\n");
while (1) {
}
}

void outl(U16_t _port, U32_t _value) {
device_cons_put_string("{outl}\n");
while (1) {
}
}

void intr_disable(void) {
	yk0a_uint32_t s;
	s = glue_lock();
	if (glue_interrupt_lock == 0) {
		yk0a_interrupt_write_mask(0);
	}
	++glue_interrupt_lock;
	glue_unlock(s);
}

void intr_enable(void) {
	yk0a_uint32_t s;
	s = glue_lock();
	if (glue_interrupt_lock != 0) {
		--glue_interrupt_lock;
		if (glue_interrupt_lock == 0) {
			yk0a_interrupt_write_mask(glue_interrupt_mask);
		}
	}
	glue_unlock(s);
}

void phys_copy(phys_bytes source, phys_bytes dest, phys_bytes count) {
	memcpy((void *)dest, (void *)source, count);
}

void alloc_segments(struct proc *rp) {
	/* ToDo: MMU */
	if (iskernelp(rp)) {
		rp->p_memmap[T].mem_vir = rp->p_memmap[T].mem_phys;
		rp->p_memmap[D].mem_vir = rp->p_memmap[D].mem_phys;
		rp->p_memmap[S].mem_vir = rp->p_memmap[S].mem_phys;
		rp->p_reg.sp -= STACK_BOOT;
	} else {
		rp->p_memmap[T].mem_vir = MINIX_PROCESS / CLICK_SIZE;
		rp->p_memmap[D].mem_vir = rp->p_memmap[T].mem_vir + rp->p_memmap[T].mem_len;
		rp->p_memmap[S].mem_vir = rp->p_memmap[D].mem_vir + rp->p_memmap[S].mem_phys - rp->p_memmap[D].mem_phys;
		if (rp->p_reg.sp == 0) {
			rp->p_reg.sp = (rp->p_memmap[S].mem_vir + rp->p_memmap[S].mem_len) * CLICK_SIZE - STACK_BOOT;
		}
	}
}

void vm_init(void) {
device_cons_put_string("{vm_init}\n");
while (1) {
}
}

void vm_map_range(u32_t base, u32_t size, u32_t offset) {
device_cons_put_string("{vm_map_range}\n");
while (1) {
}
}

phys_bytes umap_local(register struct proc *rp, int seg, vir_bytes vir_addr, vir_bytes bytes) {
	if (seg == T) {
		if (vir_addr / CLICK_SIZE >= rp->p_memmap[T].mem_vir && (vir_addr + bytes - 1) / CLICK_SIZE < rp->p_memmap[T].mem_vir + rp->p_memmap[T].mem_len) {
			return vir_addr + (rp->p_memmap[T].mem_phys - rp->p_memmap[T].mem_vir) * CLICK_SIZE;
		}
	} else {
		if (vir_addr / CLICK_SIZE >= rp->p_memmap[T].mem_vir && (vir_addr + bytes - 1) / CLICK_SIZE < rp->p_memmap[T].mem_vir + rp->p_memmap[T].mem_len) { /* ToDo: .rodata */
			return vir_addr + (rp->p_memmap[T].mem_phys - rp->p_memmap[T].mem_vir) * CLICK_SIZE;
		}
		if (vir_addr / CLICK_SIZE >= rp->p_memmap[D].mem_vir && (vir_addr + bytes - 1) / CLICK_SIZE < rp->p_memmap[S].mem_vir + rp->p_memmap[S].mem_len) {
			return vir_addr + (rp->p_memmap[D].mem_phys - rp->p_memmap[D].mem_vir) * CLICK_SIZE;
		}
	}
	return 0;
}

void cp_mess(int src, phys_clicks src_clicks, vir_bytes src_offset, phys_clicks dst_clicks, vir_bytes dst_offset) {
	message *m;
	m = (message *)(dst_clicks * CLICK_SIZE + dst_offset);
	memcpy(m, (void *)(src_clicks * CLICK_SIZE + src_offset), sizeof(message));
	m->m_source = src;
}

phys_bytes umap_remote(struct proc* rp, int seg, vir_bytes vir_addr, vir_bytes bytes) {
device_cons_put_string("{umap_remote}\n");
while (1) {
}
}

phys_bytes seg2phys(U16_t seg) {
device_cons_put_string("{seg2phys}\n");
while (1) {
}
}

void phys_memset(phys_bytes source, unsigned long pattern, phys_bytes count) {
	memset((void *)source, pattern, count);
}

vir_bytes alloc_remote_segment(u32_t *selector, segframe_t *segments, int index, phys_bytes phys, vir_bytes size, int priv) {
	/* ToDo: remote segment */
	return phys;
}

int arch_init_clock(void) {
	*(volatile yk0a_uint32_t *)DEVICE_RTC_HZ = HZ;
	return 0;
}

#if 0
clock_t read_clock(void) {
device_cons_put_string("{read_clock}\n");
while (1) {
}
}
#else /* patch */
#endif

void clock_stop(void) {
device_cons_put_string("{clock_stop}\n");
while (1) {
}
}

int intr_init(int mine) {
	return 0;
}

int intr_disabled(void) {
	return glue_interrupt_lock;
}

int intr_unmask(irq_hook_t* hook) {
	irq_actids[hook->irq] &= ~hook->id;
	if (irq_actids[hook->irq] == 0) {
		yk0a_uint32_t s;
		s = glue_lock();
		glue_interrupt_mask |= 1 << hook->irq;
		if (glue_interrupt_lock == 0) {
			yk0a_interrupt_write_mask(glue_interrupt_mask);
		}
		glue_unlock(s);
	}
	return 0;
}

int intr_mask(irq_hook_t* hook) {
	yk0a_uint32_t s;
	irq_actids[hook->irq] |= hook->id;
	s = glue_lock();
	glue_interrupt_mask &= ~(1 << hook->irq);
	if (glue_interrupt_lock == 0) {
		yk0a_interrupt_write_mask(glue_interrupt_mask);
	}
	glue_unlock(s);
	return 0;
}

void idle_task(void) {
	while (1) {
	}
}

void system_init(void) {
}

void ser_putc(char c) {
device_cons_put_string("{ser_putc}\n");
while (1) {
}
}

void arch_shutdown(int how) {
device_cons_put_string("{arch_shutdown}\n");
while (1) {
}
}

void restart(void) {
device_cons_put_string("{restart}\n");
while (1) {
}
}

void read_tsc(unsigned long *high, unsigned long *low) {
	/* ToDo: TSC */
}

int arch_init_profile_clock(u32_t freq) {
device_cons_put_string("{arch_init_profile_clock}\n");
while (1) {
}
}

void arch_stop_profile_clock(void) {
device_cons_put_string("{arch_stop_profile_clock}\n");
while (1) {
}
}

void arch_ack_profile_clock(void) {
device_cons_put_string("{arch_ack_profile_clock}\n");
while (1) {
}
}

int do_int86(message *m_ptr) {
device_cons_put_string("{do_int86}\n");
while (1) {
}
}

int do_sdevio(message *m_ptr) {
device_cons_put_string("{do_sdevio}\n");
while (1) {
}
}

int do_iopenable(message *m_ptr) {
device_cons_put_string("{do_iopenable}\n");
while (1) {
}
}

int do_readbios(message *m_ptr) {
device_cons_put_string("{do_readbios}\n");
while (1) {
}
}
